/*
 * Copyright 2020 Tadashi G. Takaoka
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "config_tms9900.h"

#include "entry_tms9900.h"
#include "table_tms9900.h"
#include "text_tms9900.h"

#include <ctype.h>
#include <string.h>

namespace libasm {
namespace tms9900 {

#define E(_insn, _name, _src, _dst)                     \
    { _insn, Entry::_flags(_src, _dst), TEXT_##_name },

static constexpr Entry TABLE_TMS9900[] PROGMEM = {
    E(0x0200, LI,   M_REG,  M_IMM)
    E(0x0220, AI,   M_REG,  M_IMM)
    E(0x0240, ANDI, M_REG,  M_IMM)
    E(0x0260, ORI,  M_REG,  M_IMM)
    E(0x0280, CI,   M_REG,  M_IMM)
    E(0x02A0, STWP, M_REG,  M_NO)
    E(0x02C0, STST, M_REG,  M_NO)
    E(0x02E0, LWPI, M_IMM,  M_NO)
    E(0x0300, LIMI, M_IMM,  M_NO)
    E(0x0340, IDLE, M_NO,   M_NO)
    E(0x0360, RSET, M_NO,   M_NO)
    E(0x0380, RTWP, M_NO,   M_NO)
    E(0x03A0, CKON, M_NO,   M_NO)
    E(0x03C0, CKOF, M_NO,   M_NO)
    E(0x03E0, LREX, M_NO,   M_NO)
    E(0x0400, BLWP, M_SRC,  M_NO)
    E(0x0440, B,    M_SRC,  M_NO)
    E(0x0480, X,    M_SRC,  M_NO)
    E(0x04C0, CLR,  M_SRC,  M_NO)
    E(0x0500, NEG,  M_SRC,  M_NO)
    E(0x0540, INV,  M_SRC,  M_NO)
    E(0x0580, INC,  M_SRC,  M_NO)
    E(0x05C0, INCT, M_SRC,  M_NO)
    E(0x0600, DEC,  M_SRC,  M_NO)
    E(0x0640, DECT, M_SRC,  M_NO)
    E(0x0680, BL,   M_SRC,  M_NO)
    E(0x06C0, SWPB, M_SRC,  M_NO)
    E(0x0700, SETO, M_SRC,  M_NO)
    E(0X0740, ABS,  M_SRC,  M_NO)
    E(0x0800, SRA,  M_REG,  M_SCNT)
    E(0x0900, SRL,  M_REG,  M_SCNT)
    E(0x0A00, SLA,  M_REG,  M_SCNT)
    E(0x0B00, SRC,  M_REG,  M_SCNT)
    E(0x1000, JMP,  M_REL,  M_NO)
    E(0x1100, JLT,  M_REL,  M_NO)
    E(0x1200, JLE,  M_REL,  M_NO)
    E(0x1300, JEQ,  M_REL,  M_NO)
    E(0x1400, JHE,  M_REL,  M_NO)
    E(0x1500, JGT,  M_REL,  M_NO)
    E(0x1600, JNE,  M_REL,  M_NO)
    E(0x1700, JNC,  M_REL,  M_NO)
    E(0x1800, JOC,  M_REL,  M_NO)
    E(0x1900, JNO,  M_REL,  M_NO)
    E(0x1A00, JL,   M_REL,  M_NO)
    E(0x1B00, JH,   M_REL,  M_NO)
    E(0x1C00, JOP,  M_REL,  M_NO)
    E(0x1D00, SBO,  M_CRU,  M_NO)
    E(0x1E00, SBZ,  M_CRU,  M_NO)
    E(0x1F00, TB,   M_CRU,  M_NO)
    E(0x2000, COC,  M_SRC,  M_DREG)
    E(0x2400, CZC,  M_SRC,  M_DREG)
    E(0x2800, XOR,  M_SRC,  M_DREG)
    E(0x2C00, XOP,  M_SRC,  M_XOP)
    E(0x3000, LDCR, M_SRC,  M_CNT)
    E(0x3400, STCR, M_SRC,  M_CNT)
    E(0x3800, MPY,  M_SRC,  M_DREG)
    E(0x3C00, DIV,  M_SRC,  M_DREG)
    E(0x4000, SZC,  M_SRC,  M_DST)
    E(0x5000, SZCB, M_SRC,  M_DST)
    E(0x6000, S,    M_SRC,  M_DST)
    E(0x7000, SB,   M_SRC,  M_DST)
    E(0x8000, C,    M_SRC,  M_DST)
    E(0x9000, CB,   M_SRC,  M_DST)
    E(0xA000, A,    M_SRC,  M_DST)
    E(0xB000, AB,   M_SRC,  M_DST)
    E(0xC000, MOV,  M_SRC,  M_DST)
    E(0xD000, MOVB, M_SRC,  M_DST)
    E(0xE000, SOC,  M_SRC,  M_DST)
    E(0xF000, SOCB, M_SRC,  M_DST)
};

static constexpr Entry TABLE_TMS9995[] PROGMEM = {
    E(0x0080, LST,  M_REG,  M_NO)
    E(0x0090, LWP,  M_REG,  M_NO)
    E(0x0180, DIVS, M_SRC,  M_NO)
    E(0x01C0, MPYS, M_SRC,  M_NO)
};

static constexpr Entry TABLE_TMS99105[] PROGMEM = {
    E(0x001C, SRAM, M_SRC2, M_CNT2)
    E(0x001D, SLAM, M_SRC2, M_CNT2)
    E(0x0029, SM,   M_SRC2, M_DST2)
    E(0x002A, AM,   M_SRC2, M_DST2)
    E(0x00B0, BLSK, M_REG,  M_IMM)
    E(0x0140, BIND, M_SRC,  M_NO)
    E(0x0C09, TMB,  M_SRC2, M_BIT2)
    E(0x0C0A, TCMB, M_SRC2, M_BIT2)
    E(0x0C0B, TSMB, M_SRC2, M_BIT2)
    E(0x0100, EVAD, M_SRC,  M_NO)
    E(0x0380, RTWP, M_RTWP, M_NO)
};

struct TableTms9900::EntryPage {
    const Entry *table;
    const Entry *end;
};

static constexpr TableTms9900::EntryPage TMS9900_PAGES[] PROGMEM = {
    { ARRAY_RANGE(TABLE_TMS9900) },
};

static constexpr TableTms9900::EntryPage TMS9995_PAGES[] PROGMEM = {
    { ARRAY_RANGE(TABLE_TMS9900) },
    { ARRAY_RANGE(TABLE_TMS9995) },
};

static constexpr TableTms9900::EntryPage TMS99105_PAGES[] PROGMEM = {
    { ARRAY_RANGE(TABLE_TMS99105) },
    { ARRAY_RANGE(TABLE_TMS9900) },
    { ARRAY_RANGE(TABLE_TMS9995) },
};

static bool acceptMode(AddrMode opr, AddrMode table) {
    if (opr == table)
        return true;
    if (opr == M_IREG || opr == M_INCR || opr == M_SYBL || opr == M_INDX)
        return table == M_SRC  || table == M_DST
            || table == M_SRC2 || table == M_DST2;
    if (opr == M_REG)
        return table == M_SRC  || table == M_DST
            || table == M_SRC2 || table == M_DST2
            || table == M_DREG || table == M_SCNT
            || table == M_CNT2;
    if (opr == M_IMM)
        return table == M_REL || table == M_SCNT
            || table == M_CNT || table == M_CNT2
            || table == M_CRU || table == M_BIT2
            || table == M_XOP || table == M_RTWP;
    return false;
}

static bool acceptModes(uint16_t flags, const Entry *entry) {
    const uint16_t table = pgm_read_word(&entry->flags);
    return acceptMode(Entry::_srcMode(flags), Entry::_srcMode(table))
        && acceptMode(Entry::_dstMode(flags), Entry::_dstMode(table));
}

Error TableTms9900::searchName(
    InsnTms9900 &insn, const EntryPage *pages, const EntryPage *end) const {
    const uint16_t flags = Entry::_flags(insn.srcMode(), insn.dstMode());
    uint8_t count = 0;
    for (const EntryPage *page = pages; page < end; page++) {
        const Entry *table =
            reinterpret_cast<Entry *>(pgm_read_ptr(&page->table));
        const Entry *end = reinterpret_cast<Entry *>(pgm_read_ptr(&page->end));
        const Entry *entry = TableBase::searchName<Entry,uint16_t>(
            insn.name(), flags, table, end, acceptModes, count);
        if (entry) {
            insn.setOpCode(pgm_read_word(&entry->opCode));
            insn.setFlags(pgm_read_word(&entry->flags));
            return OK;
        }
    }
    return count == 0 ? UNKNOWN_INSTRUCTION : OPERAND_NOT_ALLOWED;
}

static Config::opcode_t maskCode(Config::opcode_t opCode, const Entry *entry) {
    const uint16_t flags = pgm_read_word(&entry->flags);
    const AddrMode src = Entry::_srcMode(flags);
    const AddrMode dst = Entry::_dstMode(flags);
    Config::opcode_t mask = 0;
    if (src == M_REG || dst == M_REG) mask |= 0xF;
    if (src == M_SRC) mask |= 0x3F;
    if (dst == M_DST) mask |= 0xFC0;
    if (dst == M_SCNT) mask |= 0xF0;
    if (dst == M_CNT || dst == M_XOP || dst == M_DREG) mask |= 0x3C0;
    if (src == M_REL || src == M_CRU) mask |= 0xFF;
    if (src == M_RTWP) mask |= 7;
    return opCode & ~mask;
}

Error TableTms9900::searchOpCode(
    InsnTms9900 &insn, const EntryPage *pages, const EntryPage *end) const {
    for (const EntryPage *page = pages; page < end; page++) {
        const Entry *table =
            reinterpret_cast<Entry *>(pgm_read_ptr(&page->table));
        const Entry *end = reinterpret_cast<Entry *>(pgm_read_ptr(&page->end));
        const Entry *entry = TableBase::searchCode<Entry,Config::opcode_t>(
            insn.opCode(), table, end, maskCode);
        if (entry) {
            insn.setFlags(pgm_read_word(&entry->flags));
            const /*PROGMEM*/ char *name =
                reinterpret_cast<const char *>(pgm_read_ptr(&entry->name));
            insn.setName_P(name);
            return OK;
        }
    }
    return UNKNOWN_INSTRUCTION;
}

Error TableTms9900::searchName(InsnTms9900 &insn) const {
    return _error.setError(searchName(insn, _table, _end));
}

Error TableTms9900::searchOpCode(InsnTms9900 &insn) const {
    return _error.setError(searchOpCode(insn, _table, _end));
}

TableTms9900::TableTms9900() {
    setCpu(TMS9900);
}

bool TableTms9900::setCpu(CpuType cpuType) {
    _cpuType = cpuType;
    if (cpuType == TMS9900) {
        _table = ARRAY_BEGIN(TMS9900_PAGES);
        _end = ARRAY_END(TMS9900_PAGES);
        return true;
    }
    if (cpuType == TMS9995) {
        _table = ARRAY_BEGIN(TMS9995_PAGES);
        _end = ARRAY_END(TMS9995_PAGES);
        return true;
    }
    if (cpuType == TMS99105) {
        _table = ARRAY_BEGIN(TMS99105_PAGES);
        _end =  ARRAY_END(TMS99105_PAGES);
        return true;
    }
    return false;
}

const char *TableTms9900::listCpu() const {
    return TEXT_CPU_LIST;
}

const char *TableTms9900::getCpu() const {
    if (_cpuType == TMS99105) return TEXT_CPU_99105;
    return _cpuType == TMS9900 ? TEXT_CPU_9900 : TEXT_CPU_9995;
}

bool TableTms9900::setCpu(const char *cpu) {
    if (strncasecmp_P(cpu, TEXT_CPU_TMS, 3) == 0)
        cpu += 3;
    if (strcmp_P(cpu, TEXT_CPU_9900) == 0)
        return setCpu(TMS9900);
    if (strcmp_P(cpu, TEXT_CPU_9995) == 0)
        return setCpu(TMS9995);
    if (strcmp_P(cpu, TEXT_CPU_99105) == 0)
        return setCpu(TMS99105);
    return false;
}

class TableTms9900 TableTms9900;

} // namespace tms9900
} // namespace libasm

// Local Variables:
// mode: c++
// c-basic-offset: 4
// tab-width: 4
// End:
// vim: set ft=cpp et ts=4 sw=4:
